@@ -22,16 +22,19 @@ import (
 	"errors"
 	"fmt"
 	"path"
+	"strconv"
 	"strings"
 	"time"
 
-	"github.com/Azure/azure-sdk-for-go/services/compute/mgmt/2018-10-01/compute"
-	"github.com/Azure/azure-sdk-for-go/services/network/mgmt/2018-12-01/network"
+	"github.com/Azure/azure-sdk-for-go/services/compute/mgmt/2020-06-30/compute"
+	"github.com/Azure/azure-sdk-for-go/services/network/mgmt/2020-06-01/network"
+	"github.com/Azure/go-autorest/autorest"
 	"github.com/Azure/go-autorest/autorest/to"
 	machinev1 "github.com/openshift/machine-api-operator/pkg/apis/machine/v1beta1"
-	machineapierrors "github.com/openshift/machine-api-operator/pkg/controller/machine"
+	machinecontroller "github.com/openshift/machine-api-operator/pkg/controller/machine"
+	"github.com/openshift/machine-api-operator/pkg/metrics"
 	apicorev1 "k8s.io/api/core/v1"
-	"k8s.io/klog"
+	"k8s.io/klog/v2"
 	"sigs.k8s.io/cluster-api-provider-azure/pkg/apis/azureprovider/v1beta1"
 	"sigs.k8s.io/cluster-api-provider-azure/pkg/cloud/azure"
 	"sigs.k8s.io/cluster-api-provider-azure/pkg/cloud/azure/actuators"
@@ -269,7 +272,7 @@ func (s *Reconciler) Update(ctx context.Context) error {
 }
 
 func getVMState(vm compute.VirtualMachine) v1beta1.VMState {
-	if vm.ProvisioningState == nil {
+	if vm.VirtualMachineProperties == nil || vm.ProvisioningState == nil {
 		return ""
 	}
 
@@ -319,15 +322,29 @@ func (s *Reconciler) setMachineCloudProviderSpecifics(vm compute.VirtualMachine)
 
 	s.scope.Machine.Annotations[MachineInstanceStateAnnotationName] = string(getVMState(vm))
 
-	if vm.HardwareProfile != nil {
-		s.scope.Machine.Labels[MachineInstanceTypeLabelName] = string(vm.HardwareProfile.VMSize)
+	if vm.VirtualMachineProperties != nil {
+		if vm.VirtualMachineProperties.HardwareProfile != nil {
+			s.scope.Machine.Labels[MachineInstanceTypeLabelName] = string(vm.VirtualMachineProperties.HardwareProfile.VMSize)
+		}
 	}
+
 	if vm.Location != nil {
 		s.scope.Machine.Labels[MachineRegionLabelName] = *vm.Location
 	}
 	if vm.Zones != nil {
 		s.scope.Machine.Labels[MachineAZLabelName] = strings.Join(*vm.Zones, ",")
 	}
+
+	if s.scope.MachineConfig.SpotVMOptions != nil {
+		// Label on the Machine so that an MHC can select spot instances
+		s.scope.Machine.Labels[machinecontroller.MachineInterruptibleInstanceLabelName] = ""
+
+		if s.scope.Machine.Spec.Labels == nil {
+			s.scope.Machine.Spec.Labels = make(map[string]string)
+		}
+		// Label on the Spec so that it is propogated to the Node
+		s.scope.Machine.Spec.Labels[machinecontroller.MachineInterruptibleInstanceLabelName] = ""
+	}
 }
 
 // Exists checks if machine exists
@@ -350,8 +367,6 @@ func (s *Reconciler) Exists(ctx context.Context) (bool, error) {
 		return false, fmt.Errorf("returned incorrect vm interface: %T", vmInterface)
 	}
 
-	klog.Infof("Found vm for machine %s", s.scope.Name())
-
 	if s.scope.MachineConfig.UserDataSecret == nil {
 		vmExtSpec := &virtualmachineextensions.Spec{
 			Name:   "startupScript",
@@ -368,15 +383,24 @@ func (s *Reconciler) Exists(ctx context.Context) (bool, error) {
 		}
 	}
 
+	// VM States come from https://docs.microsoft.com/en-us/azure/virtual-machines/windows/states-lifecycle#provisioning-states
 	switch v1beta1.VMState(*vm.ProvisioningState) {
 	case v1beta1.VMStateSucceeded:
+		// VM was created or updated successfully
 		klog.Infof("Machine %v is running", to.String(vm.VMID))
 	case v1beta1.VMStateUpdating:
+		// Some update is being applied
 		klog.Infof("Machine %v is updating", to.String(vm.VMID))
+	case v1beta1.VMStateCreating:
+		// Creation request was accepted, VM is being created
+		klog.Infof("Machine %v is creating", to.String(vm.VMID))
 	default:
+		klog.Infof("Not found vm for machine %s", s.scope.Machine.GetName())
 		return false, nil
 	}
 
+	klog.Infof("Found vm for machine %s", s.scope.Machine.GetName())
+
 	return true, nil
 }
 
@@ -404,6 +428,11 @@ func (s *Reconciler) Delete(ctx context.Context) error {
 	}
 	err = s.disksSvc.Delete(ctx, osDiskSpec)
 	if err != nil {
+		metrics.RegisterFailedInstanceDelete(&metrics.MachineLabels{
+			Name:      s.scope.Machine.Name,
+			Namespace: s.scope.Machine.Namespace,
+			Reason:    err.Error(),
+		})
 		return fmt.Errorf("failed to delete OS disk: %w", err)
 	}
 
@@ -418,6 +447,11 @@ func (s *Reconciler) Delete(ctx context.Context) error {
 
 	err = s.networkInterfacesSvc.Delete(ctx, networkInterfaceSpec)
 	if err != nil {
+		metrics.RegisterFailedInstanceDelete(&metrics.MachineLabels{
+			Name:      s.scope.Machine.Name,
+			Namespace: s.scope.Machine.Namespace,
+			Reason:    err.Error(),
+		})
 		return fmt.Errorf("Unable to delete network interface: %w", err)
 	}
 
@@ -437,6 +471,11 @@ func (s *Reconciler) Delete(ctx context.Context) error {
 			Name: publicIPName,
 		})
 		if err != nil {
+			metrics.RegisterFailedInstanceDelete(&metrics.MachineLabels{
+				Name:      s.scope.Machine.Name,
+				Namespace: s.scope.Machine.Namespace,
+				Reason:    err.Error(),
+			})
 			return fmt.Errorf("unable to delete Public IP: %w", err)
 		}
 	}
@@ -450,7 +489,7 @@ func (s *Reconciler) getZone(ctx context.Context) (string, error) {
 
 func (s *Reconciler) createNetworkInterface(ctx context.Context, nicName string) error {
 	if s.scope.MachineConfig.Vnet == "" {
-		return machineapierrors.InvalidMachineConfiguration("MachineConfig vnet is missing on machine %s", s.scope.Machine.Name)
+		return machinecontroller.InvalidMachineConfiguration("MachineConfig vnet is missing on machine %s", s.scope.Machine.Name)
 	}
 
 	networkInterfaceSpec := &networkinterfaces.Spec{
@@ -459,7 +498,7 @@ func (s *Reconciler) createNetworkInterface(ctx context.Context, nicName string)
 	}
 
 	if s.scope.MachineConfig.Subnet == "" {
-		return machineapierrors.InvalidMachineConfiguration("MachineConfig subnet is missing on machine %s, skipping machine creation", s.scope.Machine.Name)
+		return machinecontroller.InvalidMachineConfiguration("MachineConfig subnet is missing on machine %s, skipping machine creation", s.scope.Machine.Name)
 	}
 
 	networkInterfaceSpec.SubnetName = s.scope.MachineConfig.Subnet
@@ -485,10 +524,15 @@ func (s *Reconciler) createNetworkInterface(ctx context.Context, nicName string)
 	if s.scope.MachineConfig.PublicIP {
 		publicIPName, err := azure.GenerateMachinePublicIPName(s.scope.MachineConfig.Name, s.scope.Machine.Name)
 		if err != nil {
-			return machineapierrors.InvalidMachineConfiguration("unable to create Public IP: %v", err)
+			return machinecontroller.InvalidMachineConfiguration("unable to create Public IP: %v", err)
 		}
 		err = s.publicIPSvc.CreateOrUpdate(ctx, &publicips.Spec{Name: publicIPName})
 		if err != nil {
+			metrics.RegisterFailedInstanceCreate(&metrics.MachineLabels{
+				Name:      s.scope.Machine.Name,
+				Namespace: s.scope.Machine.Namespace,
+				Reason:    err.Error(),
+			})
 			return fmt.Errorf("unable to create Public IP: %w", err)
 		}
 		networkInterfaceSpec.PublicIP = publicIPName
@@ -496,6 +540,11 @@ func (s *Reconciler) createNetworkInterface(ctx context.Context, nicName string)
 
 	err := s.networkInterfacesSvc.CreateOrUpdate(ctx, networkInterfaceSpec)
 	if err != nil {
+		metrics.RegisterFailedInstanceCreate(&metrics.MachineLabels{
+			Name:      s.scope.Machine.Name,
+			Namespace: s.scope.Machine.Namespace,
+			Reason:    err.Error(),
+		})
 		return fmt.Errorf("unable to create VM network interface: %w", err)
 	}
 
@@ -508,6 +557,11 @@ func (s *Reconciler) createVirtualMachine(ctx context.Context, nicName string) e
 		return fmt.Errorf("failed to decode ssh public key: %w", err)
 	}
 
+	priority, evictionPolicy, billingProfile, err := getSpotVMOptions(s.scope.MachineConfig.SpotVMOptions)
+	if err != nil {
+		return fmt.Errorf("failed to get Spot VM options %w", err)
+	}
+
 	vmSpec := &virtualmachines.Spec{
 		Name: s.scope.Machine.Name,
 	}
@@ -524,14 +578,17 @@ func (s *Reconciler) createVirtualMachine(ctx context.Context, nicName string) e
 		}
 
 		vmSpec = &virtualmachines.Spec{
-			Name:       s.scope.Machine.Name,
-			NICName:    nicName,
-			SSHKeyData: string(decoded),
-			Size:       s.scope.MachineConfig.VMSize,
-			OSDisk:     s.scope.MachineConfig.OSDisk,
-			Image:      s.scope.MachineConfig.Image,
-			Zone:       zone,
-			Tags:       s.scope.MachineConfig.Tags,
+			Name:           s.scope.Machine.Name,
+			NICName:        nicName,
+			SSHKeyData:     string(decoded),
+			Size:           s.scope.MachineConfig.VMSize,
+			OSDisk:         s.scope.MachineConfig.OSDisk,
+			Image:          s.scope.MachineConfig.Image,
+			Zone:           zone,
+			Tags:           s.scope.MachineConfig.Tags,
+			Priority:       priority,
+			EvictionPolicy: evictionPolicy,
+			BillingProfile: billingProfile,
 		}
 
 		if s.scope.MachineConfig.ManagedIdentity != "" {
@@ -555,6 +612,16 @@ func (s *Reconciler) createVirtualMachine(ctx context.Context, nicName string) e
 
 		err = s.virtualMachinesSvc.CreateOrUpdate(ctx, vmSpec)
 		if err != nil {
+			metrics.RegisterFailedInstanceCreate(&metrics.MachineLabels{
+				Name:      s.scope.Machine.Name,
+				Namespace: s.scope.Machine.Namespace,
+				Reason:    err.Error(),
+			})
+
+			var detailedError autorest.DetailedError
+			if errors.As(err, &detailedError) && detailedError.Message == "Failure sending request" {
+				return machinecontroller.InvalidMachineConfiguration("failure sending request for machine %s: %v", s.scope.Machine.Name, err)
+			}
 			return fmt.Errorf("failed to create or get machine: %w", err)
 		}
 	} else if err != nil {
@@ -605,3 +672,24 @@ func (s *Reconciler) getCustomUserData() (string, error) {
 
 	return base64.StdEncoding.EncodeToString(data), nil
 }
+
+func getSpotVMOptions(spotVMOptions *v1beta1.SpotVMOptions) (compute.VirtualMachinePriorityTypes, compute.VirtualMachineEvictionPolicyTypes, *compute.BillingProfile, error) {
+	// Spot VM not requested, return zero values to apply defaults
+	if spotVMOptions == nil {
+		return compute.VirtualMachinePriorityTypes(""), compute.VirtualMachineEvictionPolicyTypes(""), nil, nil
+	}
+	var billingProfile *compute.BillingProfile
+	if spotVMOptions.MaxPrice != nil && *spotVMOptions.MaxPrice != "" {
+		maxPrice, err := strconv.ParseFloat(*spotVMOptions.MaxPrice, 64)
+		if err != nil {
+			return compute.VirtualMachinePriorityTypes(""), compute.VirtualMachineEvictionPolicyTypes(""), nil, err
+		}
+		billingProfile = &compute.BillingProfile{
+			MaxPrice: &maxPrice,
+		}
+	}
+
+	// We should use deallocate eviction policy it's - "the only supported eviction policy for Single Instance Spot VMs"
+	// https://github.com/openshift/enhancements/blob/master/enhancements/machine-api/spot-instances.md#eviction-policies
+	return compute.Spot, compute.Deallocate, billingProfile, nil
+}
